/*
 * @(#) Current.java
 *
 * JOTM: Java Open Transaction Manager 
 *
 * This project is developed inside the ObjectWeb Consortium,
 * http://www.objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Contributor(s):
 * 01/11/06 Christophe Ney cney@batisseurs.com for Lutris Technologies
 *          Added ResourceManagerListener mechanism to remove ThreadData
 *          dependency.
 * 01/12/03 Dean Jennings - synchronizedMap for txXids
 *
 * --------------------------------------------------------------------------
 * $Id: Current.java,v 1.11 2003/01/13 16:16:08 jmesnil Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jotm;

import java.util.*;
import javax.naming.Referenceable;
import javax.naming.Reference;
import javax.naming.NamingException;
import javax.naming.StringRefAddr;
import javax.transaction.HeuristicMixedException;
import javax.transaction.HeuristicRollbackException;
import javax.transaction.InvalidTransactionException;
import javax.transaction.NotSupportedException;
import javax.transaction.RollbackException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.UserTransaction;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;
import org.objectweb.jonas_timer.TimerManager;
import org.objectweb.transaction.jta.TransactionManager;
import org.objectweb.transaction.jta.ResourceManagerEvent;

/**
 * <code>Current</code> is the common implementation for both
 * <code>UserTransaction</code> and <code>TransactionManager</code>.
 * <ul> 
 * <li><code>UserTransaction</code> is used by clients that want to demarcate
 * transactions themselves. It is referenceable through JNDI.</li>
 * <li><code>TransactionManager</code> is used by an application server.</li>
 * </ul>
 * 
 * <p>This object is unique in a VM, i. e. each application server has
 * <em>ONE</em> <code>Current</code> object and each client program should
 * normally issue only <em>ONE</em> lookup on JNDI.</p>
 * 
 * <p><code>Current</code> also implements <code>Referenceable</code> because of
 * JNDI.</p>
 */
public class Current implements UserTransaction, TransactionManager, Referenceable {

    // transaction associated to the thread
    // Thread specific data
    private transient static ThreadLocal threadTx = new ThreadLocal();

    // ------------------------------------------------------------------
    // Private data
    // ------------------------------------------------------------------

    // Static hashtable: Xid ---> transaction
    private transient static Map txXids = Collections.synchronizedMap(new HashMap());

    // Must be init at null, for clients that do not get UserTransaction
    private transient static Current unique = null;

    // private info	
    private transient static TimerManager timermgr = null; // local
    private transient static TransactionFactory tm = null; // local or remote

    private static final int DEFAULT_TIMEOUT = 60;
    private int transactionTimeout = DEFAULT_TIMEOUT;

    // management counter
    private transient int nb_bg_tx = 0;
    private transient int nb_rb_tx = 0;
    private transient int nb_cm_tx = 0;
    private transient int nb_to = 0;

    // ------------------------------------------------------------------
    // Constructors
    // ------------------------------------------------------------------

    /**
     * Default constructor.
     * A client does not need the TMFactory.
     */
    public Current() {

        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Current.Current()");
        }
        unique = this;
        timermgr = TimerManager.getInstance();
    }

    /**
     * Constructor for an application server. 
     * 
     * The TM factory is passed as an argument. Note that the TM factory can be
     * either local or remote.
     * 
     * @param tmfact TM Factory to use
     */
    public Current(TransactionFactory tmfact) {
        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Current.Current(TransactionFactory tmfact)");
        }
        unique = this;
        tm = tmfact;
        timermgr = TimerManager.getInstance();
    }

    /**
     * Gets the <code>TransactionManager</code> instance.
     * 
     * @return TransactionManager
     */
    public static TransactionManager getTransactionManager() {
        return (TransactionManager) unique;
    }

    // ------------------------------------------------------------------
    // UserTransaction implementation
    // ------------------------------------------------------------------

    /**
     * Creates a new transaction and associate it with the current thread.
     *
     * @exception NotSupportedException Thrown if the thread is already
     *    associated with a transaction. (nested transaction are not
     *    supported)
     *
     * @exception SystemException Thrown if the transaction manager 
     *    encounters an unexpected error condition
     */
    public void begin() throws NotSupportedException, SystemException {

        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Current.begin()");
        }
        // checks that no transaction is already associated with this thread.
        TransactionImpl tx = (TransactionImpl) threadTx.get();
        if (tx != null) {
            throw new NotSupportedException("Nested transactions not supported");
        }
        // builds a new Xid
        // - should pass servername + ip addr. (LATER)
        XidImpl otid = new XidImpl("", 0);
        // creates a new TransactionImpl object
        // - May raise SystemException
        tx = new TransactionImpl(otid, transactionTimeout);
        /// clear suspended transaction. 
        try {
            tx.doAttach(XAResource.TMJOIN);
        } catch (RollbackException e) {
            // never.
            TraceTm.error("doAttach: RollbackException");
            throw new SystemException("RollbackException in occured in begin() " + e.getMessage());
        }
        // associates transaction with current thread
        threadTx.set(tx);
        // associates this Tx with the Xid
        putTxXid(otid, tx);
        // sets a timer for the transaction
        if (timermgr != null) {
            tx.setTimer(timermgr.addTimer(tx, transactionTimeout, null, false));
        }
        // enlist all connections using the event callback this should be done in UserTransaction
        // was done in doAttach
        Stack curStack = (Stack) eventListStack.get();
        if (curStack != null) { // if there is opened connections
            List list = (List) curStack.peek();
            if (list != null) {
                for (Iterator it = list.iterator(); it.hasNext();) {
                    ((ResourceManagerEvent) it.next()).enlistConnection(tx);
                }
            } else {
                if (TraceTm.isDebugJta()) {
                    TraceTm.debugJta("Current.begin called with null list");
                }
            }
        }
    }

    /**
      * Commits the transaction associated with the current thread. When this
      * method completes, the thread becomes associated with no transaction.
      *
      * @exception RollbackException Thrown to indicate that    the transaction
      * has been rolled back rather than committed.
      *
      * @exception HeuristicMixedException Thrown to indicate that a heuristic
      * decision was made and that some relevant updates have been committed
      * while others have been rolled back.
      *
      * @exception HeuristicRollbackException Thrown to indicate that a
      * heuristic decision was made and that some relevant updates have been
      * rolled back.
      *
      * @exception SecurityException Thrown to indicate that the thread is not
      * allowed to commit the transaction.
      *
      * @exception IllegalStateException Thrown if the current thread is    not
      * associated with a transaction.
      *
      * @exception SystemException Thrown if the transaction manager encounters
      * an unexpected error condition
      */
    public void commit()
        throws
            RollbackException,
            HeuristicMixedException,
            HeuristicRollbackException,
            SecurityException,
            IllegalStateException,
            SystemException {

        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Current.commit()");
        }
        // Get Transaction
        TransactionImpl tx = (TransactionImpl) getTransaction();
        if (tx == null) {
            throw new IllegalStateException("Cannot get Transaction for commit");
        }
        // Detach Transaction from current thread
        // Must be done now (XA protocol: XA-END before XA-COMMIT)
        // TMSUCCESS should be defined in Transaction.
        tx.doDetach(XAResource.TMSUCCESS);
        // Commit Transaction. Exceptions may be raised!
        try {
            tx.commit();
        } finally {
            // Dissociates the transaction from current thread
            // Has not been done in doDetach because we need to
            // be in the transactional context for beforeCompletion
            threadTx.set(null);
        }
    }

    /**
     * Rolls back the transaction associated with the current thread. When this
     * method completes, the thread becomes associated with no transaction.
     *
     * @exception SecurityException Thrown to indicate that the thread is    not
     * allowed to roll back the transaction.
     *
     * @exception IllegalStateException Thrown if the current thread is    not
     * associated with a transaction.
     *
     * @exception SystemException Thrown if the transaction manager
     * encounters an unexpected error condition
     */
    public void rollback() throws IllegalStateException, SecurityException, SystemException {
        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Current.rollback()");
        }
        // Get Transaction
        TransactionImpl tx = (TransactionImpl) getTransaction();
        if (tx == null) {
            throw new IllegalStateException("Cannot get Transaction for rollback");
        }
        // Dissociates this Transaction from current thread
        // Must be done at this level (see JTA spec. 3.3.3)
        // TMFAIL should be defined in Transaction.
        tx.doDetach(XAResource.TMFAIL);
        threadTx.set(null);
        // Roll back the transaction. Exceptions may be raised!
        tx.rollback();
    }

    /**
     * Modify the transaction associated with the current thread such that the
     * only possible outcome of the transaction is to roll back the transaction.
     *
     * @exception IllegalStateException Thrown if the current thread is    not
     * associated with a transaction.
     *
     * @exception SystemException Thrown if the transaction manager
     * encounters an unexpected error condition
     */
    public void setRollbackOnly() throws IllegalStateException, SystemException {
        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Current.setRollbackOnly()");
        }
        // Get Transaction
        TransactionImpl tx = (TransactionImpl) getTransaction();
        if (tx == null) {
            throw new IllegalStateException("Cannot get Transaction for setRollbackOnly");
        } // Set transaction rollback only. Exceptions may be raised!
        tx.setRollbackOnly();
    }

    /**
      * Returns the status of the transaction associated with the current
      * thread.
      * 
      * @return transaction status. If no transaction is associated with the
      * current thread, this method returns the Status.NoTransaction value.
      *
      * @exception SystemException Thrown if the transaction manager
      * encounters an unexpected error condition
      */
    public int getStatus() throws SystemException {
        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Current.getStatus()");
        }
        // Get Transaction
        TransactionImpl tx = (TransactionImpl) getTransaction();
        if (tx == null) {
            return Status.STATUS_NO_TRANSACTION;
        }
        // Get TX status. Exceptions may be raised!
        return tx.getStatus();
    }

    /**
      * Modifies the value of the timeout value that is associated with the
      * transactions started by the current thread with the begin method.
      *
      *  If    an application has not called this method, the transaction
      * service uses some default value for the transaction timeout.
      *
      * @param seconds value of the timeout (in seconds). If the value is zero,
      * the transaction service restores the default value.
      *
      * @exception SystemException Thrown if the transaction manager
      * encounters an unexpected error condition
      */
    public void setTransactionTimeout(int seconds) throws SystemException {
        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Current.setTransactionTimeout(int " + seconds + ")");
        }
        if (seconds > 0) {
            transactionTimeout = seconds;
        } else {
            transactionTimeout = DEFAULT_TIMEOUT;
        }
    }

    // ------------------------------------------------------------------
    // TransactionManager implementation
    // (only the methods that are not already in UserTransaction)
    // ------------------------------------------------------------------

    /**
     * Gets the transaction object that represents the transaction context of
     * the calling thread.
     *
     * @return the Transaction object representing the transaction 
     * associated with the calling thread. If the calling thread is
     * not associated with a transaction, a null object reference
     * is returned.
     *
     * @exception SystemException Thrown if the transaction manager 
     * encounters an unexpected error condition
     */
    public Transaction getTransaction() throws SystemException {
        Transaction ret = (Transaction) threadTx.get();
        return ret;
    }

    /**
     * Resumes the transaction context association of the calling thread with
     * the transaction represented by the supplied Transaction object. When this
     * method returns, the calling thread is associated with the transaction
     * context specified. 
     * <p><em>Warning</em>: No XA start is done here. We suppose it is already
     * done after a <code>getConnection()</code>. 
     * </p>
     * The supposed programming model is: <ol>
     * 	<li><code>getConnection()</code></li>
     * 	<li>SQL code</li>
     * 	<li><code>connection.close()</code</li>
     * 	</ol>
     *
     * @param tobj The <code>Transaction</code> object that represents the
     * transaction to be resumed.
     *
     * @exception InvalidTransactionException Thrown if the parameter
     * transaction object contains an invalid transaction
     *
     * @exception IllegalStateException Thrown if the thread is already
     * associated with another transaction.
     *
     * @exception SystemException Thrown if the transaction manager
     * encounters an unexpected error condition
     */
    public void resume(Transaction tobj) throws InvalidTransactionException, IllegalStateException, SystemException {
        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Current.resume()");
        }
        // invalid Transaction
        if (tobj == null) {
            TraceTm.error("resume: null arg.");
            throw new InvalidTransactionException("resume(null) is not valid");
        }
        // Check sthat the thread is not already associated to ANOTHER transaction
        Transaction mytx = (Transaction) threadTx.get();
        if (mytx != null) {
            if (mytx.equals(tobj)) {
                if (TraceTm.isDebugJta()) {
                    TraceTm.debugJta("resume: nothing to do");
                }
                return;
            }
            TraceTm.error("resume: already associated with another transaction.");
            throw new IllegalStateException("the thread is already associated with another transaction.");
        }
        // test for type before cast
        if (!(tobj instanceof TransactionImpl)) {
            TraceTm.error("resume: non TransactionImpl arg.");
            throw new InvalidTransactionException("resume(" + tobj.getClass().getName() + ") is not valid");
        }
        // Associates this Tx with the current thread
        threadTx.set(tobj);
        // attach suspended resources
        try {
            ((TransactionImpl) tobj).doAttach(XAResource.TMRESUME);
        } catch (RollbackException e) {
            // never.
            TraceTm.error("doAttach: RollbackException");
            throw new SystemException("RollbackException in occured in begin() " + e.getMessage());
        }
    }

    /** 
      * Suspends the transaction currently associated with the calling thread
      * and return a <code>Transaction</code> object that represents the
      * transaction context being suspended. 
      * If the calling thread is not
      * associated with a transaction, the method returns
      * <code>null</code>. When this method returns, the calling thread is
      * associated with no transaction.
      <p><em>Warning</em>: No XA start is done here. We suppose it is already
      done after a <code>getConnection()</code>.
     * </p>
     * The supposed programming model is: <ol>
     * 	<li><code>getConnection()</code></li>
     * 	<li>SQL code</li>
     * 	<li><code>connection.close()</code</li>
     * 	</ol>
     * 
     * @return Transaction object representing the suspended transaction.
     *
     * @exception SystemException Thrown if the transaction manager
     * encounters an unexpected error condition
     *
     * @exception SystemException Thrown if the transaction manager
     * encounters an unexpected error condition
     */
    public Transaction suspend() throws SystemException {
        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Current.suspend()");
        }
        /// TMSUSPEND should be defined in Transaction.
        TransactionImpl tx = (TransactionImpl) threadTx.get();
        if (tx != null) {
            tx.doDetach(XAResource.TMSUSPEND);
            threadTx.set(null);
        }
        return tx;
    }

    // ------------------------------------------------------------------
    // ResourceManagerEventListener implementation
    // Should be in UserTransaction only.
    // ------------------------------------------------------------------

    /** 
      * @see org.objectweb.transaction.jta.ResourceManagerEventListener#connectionOpened(org.objectweb.transaction.jta.ResourceManagerEvent)
      */
    public void connectionOpened(ResourceManagerEvent event) {
        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Current.connectionOpened");
        }
        List list = null;
        Stack curStack = (Stack) eventListStack.get();
        // get thread local stack
        if (curStack == null) {
            // no stack yet : create one
            eventListStack.set(curStack = new Stack());
        } else { // pop list from current stack
            try {
                list = (List) curStack.pop();
            } catch (EmptyStackException e) {
                // ignore: might happen if thread used by session bean
            }
        }
        if (list == null)
            list = new Vector(1);
        // no list yet: create one
        list.add(event);
        // add event to list
        curStack.push(list);
        // push list into stack
    }

    /**
     * @see org.objectweb.transaction.jta.ResourceManagerEventListener#connectionClosed(org.objectweb.transaction.jta.ResourceManagerEvent)
     */
    public void connectionClosed(ResourceManagerEvent event) {
        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Current.connectionClosed");
        }
        // remove event from list
        removeFromCurrentStack(event);
    }

    /**
     * @see org.objectweb.transaction.jta.ResourceManagerEventListener#connectionErrorOccured(org.objectweb.transaction.jta.ResourceManagerEvent)
     */
    public void connectionErrorOccured(ResourceManagerEvent event) {
        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Current.connectionClosed");
        }
        removeFromCurrentStack(event); // remove event from list
    }

    private void removeFromCurrentStack(ResourceManagerEvent event) {
        Stack curStack = (Stack) eventListStack.get();
        // get thread local stack
        List list = (List) curStack.peek();
        // get list from stack top
        if (!list.remove(event)) {
            TraceTm.error("Current.removeRMEventFromStack called with event not in list");
        }
    }

    //
    // objectweb transaction manager implementation
    //

    // thread local variable that stores all RM Events
    ThreadLocal eventListStack = new ThreadLocal();

    /**
     * @see org.objectweb.transaction.jta.TransactionManager#pushThreadLocalRMEventList(java.util.List)
     */
    public void pushThreadLocalRMEventList(List eventList) {
        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Current.pushThreadLocalRMEventList");
        }
        Stack curStack = (Stack) eventListStack.get();
        if (curStack == null) {
            eventListStack.set(curStack = new Stack());
        }
        curStack.push(eventList);
    }

    /**
     * @see org.objectweb.transaction.jta.TransactionManager#popThreadLocalRMEventList()
     */
    public List popThreadLocalRMEventList() {
        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Current.popThreadLocalRMEventList");
        }
        Stack curStack = (Stack) eventListStack.get();
        return (List) curStack.pop();
    }

    // ------------------------------------------------------------------
    // Referenceable implementation
    // ------------------------------------------------------------------

    /**
     * Retrieves the <code>Reference</code> of this object. 
     * 
     * @return  The non-null <code>Reference</code> of this object.
     *  
     * @exception  NamingException  If a naming exception was encountered while retrieving the reference.  
     */
    public Reference getReference() throws NamingException {
        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Current.getReference()");
        }
        // create the reference
        Reference ref = new Reference(this.getClass().getName(), "org.objectweb.jotm.UserTransactionFactory", null);
        Integer i = new Integer(transactionTimeout);
        ref.add(new StringRefAddr("jonas.tm.timeout", i.toString()));
        return ref;
    }

    // ------------------------------------------------------------------
    // Other public methods
    // ------------------------------------------------------------------

    /**
     * Returns the unique instance of the class or <code>null</code> if not
     * initialized in case of plain client.
     *
     * @return The <code>Current</code> object created 
     */
    public static Current getCurrent() {
        return unique;
    }

    /**
     * Returns the TMFactory (in JTM)
     * 
     * @return The TransactionFactory
     */
    public static TransactionFactory getJTM() {
        if (tm == null) {
            TraceTm.error("Current: TMFactory is null!");
        }
        return tm;
    }

    /**
     * Sets the default timeout value
     * 
     * @param timeout timeout value (in seconds)
     */
    public void setDefaultTimeout(int timeout) {
        if (timeout != 0) {
            transactionTimeout = timeout;
        }
    }

    /**
      * Gets the default timeout value
      * 
      * @return default timeout value (in seconds)
      */
    public int getDefaultTimeout() {
        return transactionTimeout;
    }

    /**
      * Associate to the current thread a transaction represented by its
      * propagation context.
      * This is used internally by the implicit propagation of the
      * transactional context: 
      * <ul>
      * <li>in the skeleton, before calling the request (<code>isReply =
      * 	false</code>)</li>
      *  <li>in    the stub, after receiving the reply (<code>isReply =
      * 	true</code>)</li>
      * </ul>
      * 
      * @param pctx TransactionContext
      * @param isReply <code>true</code> before calling a request,
      * <code>false</code> after receiving a reply
      */
    public void setPropagationContext(TransactionContext pctx, boolean isReply) {

        if (pctx == null) { // never isReply when pctx=null
            if (TraceTm.isDebugJta()) {
                TraceTm.debugJta("Current.setPropagationContext(null) -> detach");
            }
            TransactionImpl tx = (TransactionImpl) threadTx.get();
            if (tx != null) {
                // Free TransactionImpl, if not used only!!!
                // This is to avoid memory leaks.
                if (tx.isRemovable()) {
                    forgetTx(tx.getXid());
                } // Detaches thread from any transaction
                threadTx.set(null);
            }
            return;
        }
        // Get the Transaction matching this Xid if it exists
        Xid xid = pctx.getXid();
        TransactionImpl tx = getTxXid(xid);
        if (tx == null) {
            if (!isReply) { // create a Transaction object
                if (TraceTm.isDebugJta()) {
                    TraceTm.debugJta("Current: setPropagationContext -> new Tx");
                }
                tx = new TransactionImpl(pctx);
                putTxXid(xid, tx);
            } else {
                // This may happen with Jeremie and distributed transactions with
                // no dammage.
                if (TraceTm.isDebugJta()) {
                    TraceTm.debugJta("Current.setPropagationContext: unknown tx:" + xid);
                }
            }
        } else {
            if (isReply) {
                // This thread is already associated to a transaction.
                // A proper PropagationContext exists yet, but it may have been changed
                // in case of distributed transaction.
                if (TraceTm.isDebugJta()) {
                    TraceTm.debugJta("Current.setPropagationContext: updating Xid=" + xid);
                }
                tx.updatePropagationContext(pctx);
            } else {
                if (TraceTm.isDebugJta()) {
                    TraceTm.debugJta("Current.setPropagationContext: transaction already known:" + xid);
                }
            }
        }
        // Associates this Tx with the current thread (= resume)
        // In case of Jeremy, if isReply, we are not in the correct thread, so
        // never touch threadTx in that case!!!
        if (!isReply) {
            threadTx.set(tx);
        }
    }

    /**
      * Get the propagation context associated with the current thread or null
      * if the thread is not involved in a transaction.
      */
    public TransactionContext getPropagationContext(boolean hold) {
        try {
            TransactionImpl tx = (TransactionImpl) getTransaction();
            if (tx != null) { // will return null if Tx not valid.
                return tx.getPropagationContext(hold);
            }
        } catch (SystemException e) {
            TraceTm.error("getPropagationContext system exception:", e);
        }
        return null;
    }

    /**
      * Forget all about this transaction.
      * References to <code>TransactionImpl</code> must be destroyed to allow
      * the garbage collector to free memory allocated to this transaction.
      * 
      * @param xid <code>Xid</code> of the transaction
      */
    public void forgetTx(Xid xid) {
        removeTxXid(xid);
        threadTx.set(null);
    }

    // ------------------------------------------------------------------
    // private methods
    // ------------------------------------------------------------------

    /*
     * putTxXid method.
     */
    private void putTxXid(Xid xid, TransactionImpl tx) {
        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Associate Tx to Xid:" + xid);
        }
        txXids.put(xid, tx);
    }

    /*
      * getTxXid method.
      */
    private TransactionImpl getTxXid(Xid xid) {
        TransactionImpl tx = (TransactionImpl) txXids.get(xid);
        return tx;
    }

    /*
      * removeTxXid method.
      */
    private void removeTxXid(Xid xid) {
        if (TraceTm.isDebugJta()) {
            TraceTm.debugJta("Remove Tx from Xid:" + xid);
        }
        txXids.remove(xid);
    }

    /* Management methods */

    /**
      * Returns the current number of transactions.
      * 
      * @return current number of transaction
      */
    public int getCurrentNumberOfTx() {
        return txXids.size();
    }

    /**
     * Increments number of begun transactions by one.
     */
    public synchronized void incrementBeginCounter() {
        nb_bg_tx++;
    }

    /**
     * Returns the total number of begun transactions.
     * 
     * @return total number of begun transactions
     */
    public int getTotalNumberOfBeginTx() {
        return nb_bg_tx;
    }

    /**
     * Increments the number of rolled back transaction by one.
     */
    public synchronized void incrementRollbackCounter() {
        nb_rb_tx++;
    }

    /**
     * Returns the total number of rolled back transactions.
     * 
     * @return total number of rolled back transactions
     */
    public int getTotalNumberOfRollbackTx() {
        return nb_rb_tx;
    }

    /**
     * Increments the number of of committed transactions by one.
     */
    public synchronized void incrementCommitCounter() {
        nb_cm_tx++;
    }

    /**
     * Returns the total number of committed transactions.
     *
     * @return total number of commited transaction
     */
    public int getTotalNumberOfCommitTx() {
        return nb_cm_tx;
    }

    /**
     * Resets total number of transactions.
     */
    public synchronized void resetAllTxTotalCounters() {
        nb_bg_tx = 0;
        nb_cm_tx = 0;
        nb_rb_tx = 0;
        nb_to = 0;
    }

    /**
     * Increments number of rolled back transactions due to timeout by one.
     */
    public synchronized void incrementTimeOutExpCounter() {
        nb_to++;
    }

    /**
     * Returns the total number of  rolled back transactions due to timeout.
     * 
     * @return number of rolled back transactions due to timeout
     */
    public int getTotalNumberOfTimeOutExpired() {
        return nb_to;
    }

    /**
     * Returns all counters.
     * 
     * @return an array of all counters (current tx, begun tx, committed tx,
     * rolled back tx, timeouted tx)
     */
    public synchronized Integer[] getTxCounters() {
        Integer[] result = new Integer[5];
        result[0] = new Integer(txXids.size());
        result[1] = new Integer(nb_bg_tx);
        result[2] = new Integer(nb_cm_tx);
        result[3] = new Integer(nb_rb_tx);
        result[4] = new Integer(nb_to);
        return result;
    }
}
