/*
 * @(#) Jotm.java
 *
 * JOTM: Java Open Transaction Manager
 *
 * This project is developed inside the ObjectWeb Consortium,
 * http://www.objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.objectweb.jotm;

import java.rmi.RemoteException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import javax.rmi.PortableRemoteObject;
import javax.transaction.UserTransaction;

import org.objectweb.jonas_timer.TimerManager;
import org.objectweb.transaction.jta.TransactionManager;
import org.objectweb.transaction.jta.TMService;

/**
 *  This  class represents an instance of JOTM.
 * 
 * @author jeff mesnil
*/
public class Jotm implements TMService {

    Current current = null;
    private static final String TMFACTORY_NAME = "TMFactory";
    private TransactionFactory tf = null;
    private boolean local;
    private boolean bound;

    /**
     * Public constructor for Jotm
     *
     * If Jotm is created with a <code>local</code> transaction factory which is
     * not <code>bound</code> to a registry, Jotm would be able to manage
     * transactions only inside the same JVM.
     * @param local <code>true</code> to create an instance of JOTM with a
     * local transaction factory, <code>false</code> else
     * @param bound <code>true</code> if the transaction factory is to be bound in a
     * registry, <code>false</code> else (ignored if <code>local<code> is
     * <code>false</code>)
     * 
     * @throws NamingException thrown if the transaction factory can't be bound
     * or looked up in a registry
     */
    public Jotm(boolean local, boolean bound) throws NamingException {
        TraceTm.debugTm(
            "JOTM started with a "
                + (local ? "local" : "remote")
                + " transaction factory"
                + " which is "
                + (bound ? "" : "not")
                + " bound.");
        System.setProperty(
            "javax.rmi.CORBA.PortableRemoteObjectClass",
            "org.objectweb.carol.rmi.multi.MultiPRODelegate");
        System.setProperty("java.naming.factory.initial", "org.objectweb.carol.jndi.spi.MultiOrbInitialContextFactory");
        this.local = local;
        this.bound = bound;
        if (local) {
            try {
                tf = new TransactionFactoryImpl();
            } catch (RemoteException e) {
                // should not happen: TransactionFactoryImpl throws RemoteException only
                // because it extends Remote interface
                TraceTm.error("Instanciation of TransactionFactory failed", e);
            }
            if (bound) {
                Context ictx;
                try {
                    ictx = new InitialContext();
                    ictx.rebind(TMFACTORY_NAME, tf);
                } catch (NamingException e) {
                    TraceTm.error("TransactionFactory rebind failed", e);
                    throw e;
                }
            }
        } else {
            Context ictx;
            try {
                ictx = new InitialContext();
                tf = (TransactionFactory) ictx.lookup(TMFACTORY_NAME);
            } catch (NamingException e) {
                TraceTm.error("TransactionFactory lookup failed", e);
                throw e;
            }
        }
        current = new Current(tf);
    }

    /**
     * @see org.objectweb.transaction.jta.TMService#getTransactionManager()
     */
    public TransactionManager getTransactionManager() {
        return (TransactionManager) current;
    }

    /**
     * @see org.objectweb.transaction.jta.TMService#getUserTransaction()
     */
    public UserTransaction getUserTransaction() {
        return (UserTransaction) current;
    }
    /**
     * @see org.objectweb.transaction.jta.TMService#stop()
     */
    public void stop() {
        // unexport remote objects
        if (local) {
            try {
                InitialContext ictx = new InitialContext();
                ictx.unbind(TMFACTORY_NAME);
                PortableRemoteObject.unexportObject(tf);
            } catch (Exception e) {
                //ignore
            }
        }
        tf = null;
        // stop the timer manager
        TimerManager.stop(false);
    }
}
