/*
 * @(#) TraceTm.java 
 *
 * JOTM: Java Open Transaction Manager 
 *
 * This project is developed inside the ObjectWeb Consortium,
 * http://www.objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: TraceTm.java,v 1.7 2003/01/14 15:10:11 jmesnil Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jotm;

import java.util.Properties;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.HandlerFactory;
import org.objectweb.util.monolog.api.LevelFactory;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.file.monolog.PropertiesConfAccess;

public class TraceTm {

    /** prefix used to identify JOTM loggers */
    public static final String prefix = "org.objectweb.jotm";

    // tmLogger and jtaLogger are children of jotmLogger
    // jotmLogger is used to log both verbose and error messages
    protected static Logger jotmLogger = null;
    protected static Logger tmLogger = null;
    protected static Logger jtaLogger = null;

    /**
    * Configure the log for JOTM.
    * Log configuration is stored in a property file, <code>trace.properties</code>,
    * which should be available from the classpath.
    */
    public static void configure() {
        Properties props = new Properties();
        try {
            props.load(Thread.currentThread().getContextClassLoader().getResourceAsStream("trace.properties"));
        } catch (Exception e) {
            System.err.println("Configuration file for log not found. Traces are disabled: " + e);
            return;
        }
        // Instanciate the LoggerFactory
        String lfClassName = props.getProperty("log.config.classname", null);
        if (lfClassName == null) {
            System.err.println("Malformed log configuration file: log.config.classname not available");
            return;
        }
        try {
            LoggerFactory lf = (LoggerFactory) Class.forName(lfClassName).newInstance();
            // Configure the LoggerFactory with the properties
            PropertiesConfAccess.load(props, lf, (HandlerFactory) lf, (LevelFactory) lf);
            TraceTm.configure(lf);
        } catch (Exception e) {
            System.err.println("Logs are disabled:" + e);
        }
    }

    /**
    * Configure the log for JOTM
    * 
    * @param <code>lf</code> the LoggerFactory
    */
    public static void configure(LoggerFactory lf) {
        jotmLogger = lf.getLogger(prefix);
        tmLogger = lf.getLogger(prefix + ".tm");
        jtaLogger = lf.getLogger(prefix + ".jta");
    }

    /**
     * Log a verbose message
     * 
     * @param <code>msg</code> verbose message
     */
    public static void verbose(String msg) {
        if (jotmLogger != null) {
            jotmLogger.log(BasicLevel.INFO, msg);
        }
    }

    /**
     * Log an error message.
     * 
     * @param <code>msg </code> error message
     */
    public static void error(String msg) {
        if (jotmLogger != null) {
            jotmLogger.log(BasicLevel.ERROR, msg);
        }
    }

    /**
     * Log an error message and a stack trace from a Throwable object.
     * 
     * @param <code>msg</code> error message
     * @param <code>th</code> Throwable object
     */
    public static void error(String msg, Throwable th) {
        if (jotmLogger != null) {
            jotmLogger.log(BasicLevel.ERROR, msg, th);
        }
    }

    /**
     * Test if TM debug messages are logged.
     * 
     * @return boolean  <code>true</code> if TM debug messages are logged,
     * <code>false</code> otherwise
     */
    static public boolean isDebugTm() {
        return (tmLogger != null) && tmLogger.isLoggable(BasicLevel.DEBUG);
    }

    /**
     * Log a TM debug message.
     * 
     * @param <code>msg</code> TM debug message
     */
    public static void debugTm(String msg) {
        if (tmLogger != null) {
            tmLogger.log(BasicLevel.DEBUG, msg);
        }
    }

    /**
      * Test if JTA debug messages are logged.
      *
      * @return boolean  <code>true</code> if JTA debug messages are logged,
      * <code>false</code> otherwise
      */
    static public boolean isDebugJta() {
        return (jtaLogger != null) && jtaLogger.isLoggable(BasicLevel.DEBUG);
    }

    /**
     * Log a JTA debug message.
     *
     * @param <code>msg</code> JTA debug message
     */
    public static void debugJta(String msg) {
        if (jtaLogger != null) {
            jtaLogger.log(BasicLevel.DEBUG, msg);
        }
    }
}
